@@ -27,16 +27,29 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
+#include <signal.h>
+#include <sys/time.h>
 #include <sys/stat.h>
 #include "alsactl.h"
+#include "os_compat.h"
 
-static int state_lock_(const char *file, int lock, int timeout, int _fd)
+#define PATH_SIZE 512
+
+static int alarm_flag;
+
+static void signal_handler_alarm(int sig)
+{
+	alarm_flag = 1;
+}
+
+static int state_lock_(const char *lock_file, int lock, int timeout, int _fd)
 {
 	int fd = -1, err = 0;
 	struct flock lck;
 	struct stat st;
 	char lcktxt[14];
-	char *nfile = lockfile;
+	struct sigaction sig_alarm, sig_alarm_orig;
+	struct itimerval itv;
 
 	if (do_lock <= 0)
 		return 0;
@@ -53,21 +66,25 @@ static int state_lock_(const char *file, int lock, int timeout, int _fd)
 		fd = _fd;
 	}
 	while (fd < 0 && timeout-- > 0) {
-		fd = open(nfile, O_RDWR);
+		fd = open(lock_file, O_RDWR);
 		if (!lock && fd < 0) {
 			err = -EIO;
 			goto out;
 		}
 		if (fd < 0) {
-			fd = open(nfile, O_RDWR|O_CREAT|O_EXCL, 0644);
+			fd = open(lock_file, O_RDWR|O_CREAT|O_EXCL, 0644);
 			if (fd < 0) {
 				if (errno == EBUSY || errno == EAGAIN) {
 					sleep(1);
-					timeout--;
-				} else {
-					err = -errno;
-					goto out;
+					continue;
 				}
+				if (errno == EEXIST) {
+					fd = open(lock_file, O_RDWR);
+					if (fd >= 0)
+						break;
+				}
+				err = -errno;
+				goto out;
 			}
 		}
 	}
@@ -89,15 +106,31 @@ static int state_lock_(const char *file, int lock, int timeout, int _fd)
 			goto out;
 		}
 	}
-	while (timeout > 0) {
-		if (fcntl(fd, F_SETLK, &lck) < 0) {
-			sleep(1);
-			timeout--;
-		} else {
-			break;
-		}
+	alarm_flag = 0;
+	memset(&sig_alarm, 0, sizeof(sig_alarm));
+	sigemptyset(&sig_alarm.sa_mask);
+	sig_alarm.sa_handler = signal_handler_alarm;
+	if (sigaction(SIGALRM, &sig_alarm, &sig_alarm_orig) < 0) {
+		err = -ENXIO;
+		goto out;
 	}
-	if (timeout <= 0) {
+	memset(&itv, 0, sizeof(itv));
+	itv.it_value.tv_sec = timeout;
+	if (setitimer(ITIMER_REAL, &itv, NULL) < 0) {
+		err = -ENXIO;
+		sigaction(SIGALRM, &sig_alarm_orig, NULL);
+		goto out;
+	}
+	while (alarm_flag == 0) {
+		if (fcntl(fd, F_SETLKW, &lck) == 0)
+			break;
+		if (errno == EAGAIN || errno == ERESTART)
+			continue;
+	}
+	memset(&itv, 0, sizeof(itv));
+	setitimer(ITIMER_REAL, &itv, NULL);
+	sigaction(SIGALRM, &sig_alarm_orig, NULL);
+	if (alarm_flag) {
 		err = -EBUSY;
 		goto out;
 	}
@@ -116,11 +149,21 @@ out:
 	return err;
 }
 
+static void state_lock_file(char *buf, size_t buflen)
+{
+	if (lockfile[0] == '/')
+		snprintf(buf, buflen, "%s", lockfile);
+	else
+		snprintf(buf, buflen, "%s/%s", lockpath, lockfile);
+}
+
 int state_lock(const char *file, int timeout)
 {
+	char fn[PATH_SIZE];
 	int err;
 
-	err = state_lock_(file, 1, timeout, -1);
+	state_lock_file(fn, sizeof(fn));
+	err = state_lock_(fn, 1, timeout, -1);
 	if (err < 0)
 		error("file %s lock error: %s", file, strerror(-err));
 	return err;
@@ -128,10 +171,41 @@ int state_lock(const char *file, int timeout)
 
 int state_unlock(int _fd, const char *file)
 {
+	char fn[PATH_SIZE];
 	int err;
 
-	err = state_lock_(file, 0, 10, _fd);
+	state_lock_file(fn, sizeof(fn));
+	err = state_lock_(fn, 0, 10, _fd);
 	if (err < 0)
 		error("file %s unlock error: %s", file, strerror(-err));
 	return err;
 }
+
+static void card_lock_file(char *buf, size_t buflen, int card_number)
+{
+	snprintf(buf, buflen, "%s/card%i.lock", lockpath, card_number);
+}
+
+int card_lock(int card_number, int timeout)
+{
+	char fn[PATH_SIZE];
+	int err;
+
+	card_lock_file(fn, sizeof(fn), card_number);
+	err = state_lock_(fn, 1, timeout, -1);
+	if (err < 0)
+		error("card %d lock error: %s", card_number, strerror(-err));
+	return err;
+}
+
+int card_unlock(int _fd, int card_number)
+{
+	char fn[PATH_SIZE];
+	int err;
+
+	card_lock_file(fn, sizeof(fn), card_number);
+	err = state_lock_(fn, 0, 10, _fd);
+	if (err < 0)
+		error("card %d unlock error: %s", card_number, strerror(-err));
+	return err;
+}
