@@ -331,6 +331,8 @@ _("Usage: alsabat [-options]...\n"
 "      --log=#            file that both stdout and strerr redirecting to\n"
 "      --file=#           file for playback\n"
 "      --saveplay=#       file that storing playback content, for debug\n"
+"      --readcapture=#    file with previously captured content.  File data\n"
+"                         is used for analysis instead of capturing it.\n"
 "      --local            internal loop, set to bypass pcm hardware devices\n"
 "      --standalone       standalone mode, to bypass analysis\n"
 "      --roundtriplatency round trip latency mode\n"
@@ -397,6 +399,7 @@ static void parse_arguments(struct bat *bat, int argc, char *argv[])
 		{"roundtriplatency", 0, 0, OPT_ROUNDTRIPLATENCY},
 		{"snr-db",   1, 0, OPT_SNRTHD_DB},
 		{"snr-pc",   1, 0, OPT_SNRTHD_PC},
+		{"readcapture", 1, 0, OPT_READCAPTURE},
 		{0, 0, 0, 0}
 	};
 
@@ -412,6 +415,11 @@ static void parse_arguments(struct bat *bat, int argc, char *argv[])
 		case OPT_SAVEPLAY:
 			bat->debugplay = optarg;
 			break;
+		case OPT_READCAPTURE:
+			bat->capturefile = optarg;
+			bat->capture.mode = MODE_ANALYZE_ONLY;
+			bat->playback.mode = MODE_ANALYZE_ONLY;
+			break;
 		case OPT_LOCAL:
 			bat->local = true;
 			break;
@@ -538,8 +546,8 @@ static int bat_init(struct bat *bat)
 	if (bat->logarg) {
 		bat->log = NULL;
 		bat->log = fopen(bat->logarg, "wb");
-		err = -errno;
 		if (bat->log == NULL) {
+			err = -errno;
 			fprintf(bat->err, _("Cannot open file: %s %d\n"),
 					bat->logarg, err);
 			return err;
@@ -564,17 +572,16 @@ static int bat_init(struct bat *bat)
 	} else {
 		/* create temp file for sound record and analysis */
 		fd = mkstemp(name);
-		err = -errno;
 		if (fd == -1) {
+			err = -errno;
 			fprintf(bat->err, _("Fail to create record file: %d\n"),
 					err);
 			return err;
 		}
 		/* store file name which is dynamically created */
 		bat->capture.file = strdup(name);
-		err = -errno;
 		if (bat->capture.file == NULL)
-			return err;
+			return -ENOMEM;
 		/* close temp file */
 		close(fd);
 	}
@@ -602,8 +609,8 @@ static int bat_init(struct bat *bat)
 		}
 	} else {
 		bat->fp = fopen(bat->playback.file, "rb");
-		err = -errno;
 		if (bat->fp == NULL) {
+			err = -errno;
 			fprintf(bat->err, _("Cannot open file: %s %d\n"),
 					bat->playback.file, err);
 			return err;
@@ -708,6 +715,15 @@ int main(int argc, char *argv[])
 		goto analyze;
 	}
 
+	if (bat.capture.mode == MODE_ANALYZE_ONLY && bat.capturefile) {
+		bat.capture.file = strdup(bat.capturefile);
+		fprintf(bat.log,
+			_("Using data from file %s for analysis\n"),
+			bat.capture.file);
+		fprintf(bat.log, _("Skipping playback and capture\n"));
+		goto analyze;
+	}
+
 	/* loopback thread: playback and capture in a loop */
 	if (bat.local == false)
 		test_loopback(&bat);
